"use client";

import {
  toSocket,
  WebSocketMessageReader,
  WebSocketMessageWriter,
} from "vscode-ws-jsonrpc";
import {
  CloseAction,
  ErrorAction,
  MessageTransports,
} from "vscode-languageclient";
import "@fontsource-variable/fira-code";
import normalizeUrl from "normalize-url";
import { createHighlighter } from "shiki";
import type { editor } from "monaco-editor";
import { shikiToMonaco } from "@shikijs/monaco";
import { useEffect, useRef, useMemo } from "react";
import { useCodeEditorStore } from "@/store/codeEditorStore";
import { MonacoLanguageClient } from "monaco-languageclient";
import MonacoEditor, { type Monaco } from "@monaco-editor/react";
import { initServices } from "monaco-languageclient/vscode/services";

const ReconnectingWebSocket = require("reconnecting-websocket");

const ADDITIONAL_THEMES = [
  "andromeeda",
  "aurora-x",
  "ayu-dark",
  "catppuccin-frappe",
  "catppuccin-latte",
  "catppuccin-macchiato",
  "catppuccin-mocha",
  "dark-plus",
  "dracula",
  "dracula-soft",
  "one-dark-pro",
  "vitesse-dark",
  "vitesse-light",
];

const ADDITIONAL_LANGUAGES = [
  "c",
  "cpp",
  "java",
  "python",
  "typescript",
] as const satisfies Parameters<typeof createHighlighter>[0]["langs"];

export function CodeEditor() {
  const monacoRef = useRef<Monaco | null>(null);
  const editorRef = useRef<editor.IStandaloneCodeEditor | null>(null);
  const { lang, path, theme, value, isMinimap, isLigature } =
    useCodeEditorStore();

  useEffect(() => {
    if (monacoRef.current && editorRef.current) {
      monacoRef.current.editor.setTheme(theme);
    }
  }, [theme]);

  const options = useMemo(
    () =>
      ({
        minimap: { enabled: isMinimap, size: "fit" },
        fontSize: 14,
        fontFamily: "Fira Code Variable, monospace",
        tabSize: 4,
        showFoldingControls: "always",
        fontLigatures: isLigature,
        automaticLayout: true,
        autoIndent: "full",
        guides: {
          bracketPairs: true,
          indentation: true,
        },
      } as editor.IStandaloneEditorConstructionOptions),
    [isMinimap, isLigature]
  );

  const createLanguageClient = (
    transports: MessageTransports
  ): MonacoLanguageClient => {
    return new MonacoLanguageClient({
      name: "LSP Language Client",
      clientOptions: {
        documentSelector: ADDITIONAL_LANGUAGES,
        errorHandler: {
          error: () => ({ action: ErrorAction.Continue }),
          closed: () => ({ action: CloseAction.DoNotRestart }),
        },
      },
      connectionProvider: {
        get: () => {
          return Promise.resolve(transports);
        },
      },
    });
  };

  useEffect(() => {
    if (!monacoRef.current) return;
    const url = normalizeUrl(`ws://172.20.0.13:3000`);
    const socketOptions = {
      maxReconnectionDelay: 10000,
      minReconnectionDelay: 1000,
      reconnectionDelayGrowFactor: 1.3,
      connectionTimeout: 10000,
      maxRetries: Infinity,
      debug: true,
    };
    const webSocket: WebSocket = new ReconnectingWebSocket.default(
      url,
      [],
      socketOptions
    );

    webSocket.onopen = () => {
      console.log("WebSocket connection opened");
      const socket = toSocket(webSocket);
      const reader = new WebSocketMessageReader(socket);
      const writer = new WebSocketMessageWriter(socket);
      const languageClient = createLanguageClient({ reader, writer });
      languageClient.start();
      reader.onClose(() => languageClient.stop());
    };

    webSocket.onerror = (error) => {
      console.error("WebSocket error:", error);
    };

    webSocket.onclose = (event) => {
      console.log("WebSocket connection closed:", event);
    };

    return () => {
      webSocket.close();
    };
  }, [monacoRef]);

  const handleEditorMount = async (
    editor: editor.IStandaloneCodeEditor,
    monaco: Monaco
  ) => {
    editorRef.current = editor;
    monacoRef.current = monaco;

    for (const lang of ADDITIONAL_LANGUAGES) {
      monacoRef.current?.languages.register({ id: lang });
    }

    const highlighter = await createHighlighter({
      themes: ADDITIONAL_THEMES,
      langs: ADDITIONAL_LANGUAGES,
    });

    shikiToMonaco(highlighter, monacoRef.current);

    await initServices({});
  };

  return (
    <MonacoEditor
      defaultLanguage={lang}
      theme={theme}
      path={path}
      options={options}
      defaultValue={value}
      onMount={handleEditorMount}
    />
  );
}
